'''
Used to edit files and folders.
'''

import shutil
import os

def test_folder(location: str, location_is_file: bool = True) -> None:
    '''
    Looks for a folder and creates it if it does not already exist.
    '''
    folder = location
    if location_is_file:
        if location.rfind("/") > -1:
            folder = location[0:location.rfind("/")]

    folder_hierarchy = folder.split("/")
    # Check top level folder
    folder = folder_hierarchy[0]
    if not os.path.exists(folder):
        os.makedirs(folder)
    # Check subfolders
    for i in range(len(folder_hierarchy)):
        folder = "/".join(folder_hierarchy[0:i+1])
        if not os.path.exists(folder):
            os.makedirs(folder)

def delete_contents(location: str, del_files: bool = True, del_folders: bool = True) -> None:
    '''
    Location must be a folder.\n
    Deletes all files and folders inside the specified folder, unless the optional parameters specify otherwise.
    '''
    for root, dirs, files in os.walk(location):
        if del_files:
            for f in files:
                os.unlink(os.path.join(root, f))
        if del_folders:
            for d in dirs:
                shutil.rmtree(os.path.join(root, d))

def delete(location: str) -> None:
    '''
    Deletes a file or folder.
    '''
    if os.path.isfile(location):
        os.remove(location)
    elif os.path.isdir(location):
        shutil.rmtree(location)
    else:
        print(f"Warning: {location} file/folder not found")

def copy(source: str, destination: str) -> None:
    '''
    Copies a file from source to destination.
    '''
    test_folder(destination)
    shutil.copy(source, destination)

def batch_copy(source_folder: str, dest_folder: str, files: str|list) -> None:
    '''
    Copies a list of files from source_folder to dest_folder.
    '''
    if isinstance(files, str):
        files = [files]
    for f in files:
        copy(f"{source_folder}/{f}", f"{dest_folder}/{f}")

def move(source: str, destination: str) -> None:
    '''
    Moves a file from source to destination.
    '''
    test_folder(destination)
    shutil.move(source, destination)

def read(location: str, codec, split = True) -> list[str]: # utf-8-sig
    '''
    Opens a file and reads the contents into a list of strings (or a single string in a len 1 list, if split is False).
    The file is then closed.\n
    codec can be used to change the encoding with which the file is read. Leave as "" for default.
    '''
    with open(location, "r", encoding = codec) as f:
        data = f.read()
    if split:
        data = data.split("\n")
    else:
        data = [data]
    return data

def write(location: str, data: str, codec) -> None: # cp1252 (<- ANSI for Windows)
    '''
    Creates a file, or overwrites it if it already exists, and writes the data to it.\n
    codec can be used to change the encoding with which the file is read. Leave as "" for default.
    '''
    test_folder(location)
    try:
        with open(location, "x", encoding = codec) as f:
            f.write(data)
    except FileExistsError as e:
        print(e)
        with open(location, "w", encoding = codec) as f:
            f.write(data)

def append(location: str, data: str, codec) -> None:
    '''
    Appends to an existing file, or creates a new file if it does not already exist, and writes the data to it.\n
    '''
    with open(location, "a", encoding = codec) as f:
        f.write(data)

def list_files(location: str) -> list[str]:
    '''
    Returns a list of all files in the specified folder.
    '''
    # Code found at https://stackoverflow.com/questions/3207219/how-do-i-list-all-files-of-a-directory
    return [f for f in os.listdir(location) if os.path.isfile(os.path.join(location, f))]



'''
Used to access SQL files. This is a wrapper for Sqlite3.
'''
from sqlite3 import Connection
from sqlite3 import connect as conn
from sqlite3 import Error

class Table:
    '''Initialises a table, which is an object that is used to interface with a specific table within the sql database.'''
    def __init__(self, connection: Connection, table_name: str, fields: list[str]) -> None:
        self._connection = connection
        self._table_name = table_name
        table_formula = f"CREATE TABLE {self._table_name} ("+", ".join(fields)+")"
        try:
            self._connection.cursor().execute(table_formula)
            self._connection.commit()
        except Error:
            pass

    def list_record(self, conditions: str = "", columns: str = "*") -> list[list]:
        '''Lists all recs where the condition is true.'''
        rec_bank = []
        for row in self._connection.cursor().execute(f"""SELECT {columns} FROM {self._table_name} {conditions}"""):
            rec_bank.append(row)
        return rec_bank

    def update_record(self, key: str, value, conditions: str = "") -> bool:
        '''Attempts to update a record. Returns True if an error occurred, and False otherwise.'''
        try:
            self._connection.cursor().execute(f"""UPDATE {self._table_name} SET {key} = {value} {conditions}""")
            self._connection.commit()
        except Error:
            return True
        return False

    def add_record(self, *values) -> bool:
        '''Attempts to add a record. Returns True if an error occurred, and False otherwise.'''
        try:
            self._connection.cursor().execute(f'''INSERT INTO {self._table_name} VALUES ({", ".join(values)})''')
            self._connection.commit()
        except Error:
            return True
        return False


def connect(database_name: str) -> Connection:
    '''Returns the database connection, which needs to be passed into most other functions from this file.'''
    connection = ""
    while connection == "":
        try:
            connection = conn(database_name)
        except Error:
            pass
    return connection

def close_connection(connection: Connection) -> None:
    '''Commits any unsaved changes and closes the connection to the database.'''
    connection.commit()
    connection.close()



'''
Used for various common tasks.
'''
from dataclasses import dataclass

@dataclass
class Colours:
    '''Holds colours for the terminal.'''
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

def empty(a: int|list|str) -> bool:
    '''
    Returns True if a is empty ("", [], "-", etc.).'''
    if isinstance(a, list):
        if len(a) == 0:
            return True
        for i in a:
            if not empty(i):
                return False
        return True
    if isinstance(a, str):
        return len(a) == 0 or a in ["-", " ", "."]
    if isinstance(a, int):
        return a == 0
    return False

def latinise(word: str):
    '''
    Converts a string to latin characters.
    '''
    # āçēīłṉōṟşūƶĀÇĒĪŁṈŌṞŞŪƵ
    word = word.replace("ā", "a" )
    word = word.replace("ç", "ch")
    word = word.replace("ē", "e" )
    word = word.replace("ī", "i" )
    word = word.replace("ł", "w" )
    word = word.replace("ṉ", "ny")
    word = word.replace("ō", "o" )
    word = word.replace("ṟ", "rw")
    word = word.replace("ş", "sh")
    word = word.replace("ū", "u" )
    word = word.replace("ƶ", "zh")
    word = word.replace("Ā", "A" )
    word = word.replace("Ç", "Ch")
    word = word.replace("Ē", "E" )
    word = word.replace("Ī", "I" )
    word = word.replace("Ł", "W" )
    word = word.replace("Ṉ", "Ny")
    word = word.replace("Ō", "O" )
    word = word.replace("Ṟ", "Rw")
    word = word.replace("Ş", "Sh")
    word = word.replace("Ū", "U" )
    word = word.replace("Ƶ", "Zh")
    return word



'''
Used to retrieve data from a Google Sheets spreadsheet.
'''

import os.path
from google.auth.transport.requests import Request
from google.auth.exceptions import RefreshError
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build
from googleapiclient.errors import HttpError
from googleapiclient.discovery import Resource

def _get_token_and_scopes(token_folder:str):
    # If modifying these scopes, delete the file token.json.
    scopes = ["https://www.googleapis.com/auth/spreadsheets.readonly"]
    token_location = f"{token_folder}/token.json"
    credentials_location = f"{token_folder}/credentials.json"
    return [scopes, token_location, credentials_location]

def _remove_str_list(data: list[list[str]], remove_str_list: list[str]) -> list[list[str]]:
    '''
    Removes any unwanted strings from data.
    '''
    if "\r" not in remove_str_list:
        remove_str_list.append("\r")
    for i in data:
        for j in i:
            for k in remove_str_list:
                j = j.replace(k, "")
    return data

def call_spreadsheet_api(token_folder:str) -> list[list[str]]:
    '''
    Calls the Google Sheets API to retrieve the relevant range from the given spreadsheet.\n
    Data is returned as a list of lists, with each inner list being one row of the spreadsheet.
    '''
    scopes, token_location, credentials_location = _get_token_and_scopes(token_folder)
    creds = None
    if os.path.exists(token_location):
        creds = Credentials.from_authorized_user_file(token_location, scopes)
    # If there are no (valid) credentials available, let the user log in.
    if not creds or not creds.valid:
        try:
            if creds and creds.expired and creds.refresh_token:
                creds.refresh(Request())
            else:
                flow = InstalledAppFlow.from_client_secrets_file(credentials_location, scopes)
                creds = flow.run_local_server(port=0)
        except RefreshError as e:
            print(e)
            flow = InstalledAppFlow.from_client_secrets_file(credentials_location, scopes)
            creds = flow.run_local_server(port=0)
        # Save the credentials for the next run
        with open(token_location, "w", encoding="utf-8") as token:
            token.write(creds.to_json())
    return creds


def retrieve_range(token_folder:str, spreadsheet_id: str, range_name: str, remove_str_list: list | None = None) -> list[list[str]]:
    '''
    Retrieves the relevant range from the given spreadsheet.\n
    Data is returned as a list of lists, with each inner list being one row of the spreadsheet.
    '''
    if remove_str_list is None: # Recommended by PyLint
        remove_str_list = []
    creds = call_spreadsheet_api(token_folder)

    data = []
    try:
        service: Resource = build("sheets", "v4", credentials=creds) # build is a general-purpose function that returns a dynamic object, so can't be type-checked

        # Call the Sheets API
        sheet = service.spreadsheets() # pylint: disable=no-member
        result = sheet.values().get(spreadsheetId=spreadsheet_id, range=range_name).execute()
        values = result.get("values", [])

        if not values:
            print("No data found.")
            return data

        for row in values:
            data.append(row)
    except HttpError as err:
        print(err)
    data = _remove_str_list(data, remove_str_list)
    return data

def retrieve_range_index(token_folder:str, spreadsheet_id: str, range_name: str, remove_str_list: list | None = None) -> tuple[list[list[str]], dict[str, int]]:
    '''
    Identical to retrieve_range, except it also creates an index of the location of all data on the first row (which is assumed to be the header row). The first row is not removed.
    '''
    if remove_str_list is None:
        remove_str_list = []
    range_list = retrieve_range(token_folder, spreadsheet_id, range_name, remove_str_list)
    index: dict[str, int] = {}
    for i in range_list[0]: # Populate index
        index[i] = range_list[0].index(i)
    return range_list, index

def get_note(token_folder:str, spreadsheet_id: str, row: str, col: str):
    '''
    Retrieves the note from the specified cell of a spreadsheet.\n
    Data is returned as a list of lists, with each inner list being one row of the spreadsheet.
    '''
    coord = chr(row+64)+str(col)
    print(coord)

    creds = call_spreadsheet_api(token_folder)
    
    try:
        service: Resource = build("sheets", "v4", credentials=creds) # build is a general-purpose function that returns a dynamic object, so can't be type-checked

        # Call the Sheets API
        sheet = service.spreadsheets().get(spreadsheet_id) # pylint: disable=no-member
        result = sheet.data().rowData().values().note()
#        result = sheet.get(spreadsheetId=spreadsheet_id, fields="sheets/data/rowData/values/note").execute()
        values = result.execute()#.get("sheets/data/rowData/values/note")#("values", [])

        if not values:
            print("No data found.")
    except HttpError as err:
        print(err)
    return values
